@@ -4,21 +4,19 @@
 #
 # This file is part of cloud-init. See LICENSE file for license information.
 
-__VERSION__ = "20.2"
-_PACKAGED_VERSION = '20.2-45-g5f7825e2-0ubuntu1~19.10.1'
+__VERSION__ = "25.3"
+_PACKAGED_VERSION = "25.3-0ubuntu1~22.04.1"
 
 FEATURES = [
     # supports network config version 1
-    'NETWORK_CONFIG_V1',
+    "NETWORK_CONFIG_V1",
     # supports network config version 2 (netplan)
-    'NETWORK_CONFIG_V2',
+    "NETWORK_CONFIG_V2",
 ]
 
 
 def version_string():
     """Extract a version string from cloud-init."""
-    if not _PACKAGED_VERSION.startswith('@@'):
+    if not _PACKAGED_VERSION.startswith("@@"):
         return _PACKAGED_VERSION
     return __VERSION__
-
-# vi: ts=4 expandtab
